/* 
 * 1. Once complete create intermediate versions of the code which
 *    tutorial students can complete as hands on exercises.
 *    a. create pup routines
 *    b. object complication with persistent heap data, then pup it
 */


#include "LB_Test.decl.h"

/*readonly*/ CProxy_Main mainProxy;
/*readonly*/ int num_chare_blocks;
/*readonly*/ int workWeight;
/*readonly*/ int total_iterations;
/*readonly*/ int balance_iteration;
/*readonly */ int arraySize;
class Main : public CBase_Main
{
public:
  CProxy_LB_Test arrayProxy;
  double timestamp;
  double workStartTimestamp;
  double totalChareWorkTime;
  
  Main(CkArgMsg* m) {
    int totalArraySize;
    CkAssert(CkMyPe() == 0);
    if ((m->argc < 3) || (m->argc > 5)) {
      CkPrintf("Usage: %s <num chare blocks/pe> <total iter> [work weight] [totalArraySize]\n", 
	       m->argv[0]);
      CkAbort("Abort");
    }
    num_chare_blocks = atoi(m->argv[1]);
    total_iterations = atoi(m->argv[2]);
    workWeight = 1;
    if (m->argc >= 4) {
      workWeight = atoi(m->argv[3]);
    }
    totalArraySize=14000;
    if (m->argc >= 5) {
      totalArraySize = atoi(m->argv[4]);
    }
    arraySize=totalArraySize/CkNumPes();
    if(arraySize<=1)
      {
	CkAbort("you don't have enough work for this many processors\n");
      }
    timestamp = CkWallTimer();
    totalChareWorkTime = 0.0;
    balance_iteration = total_iterations/4;
    balance_iteration = (balance_iteration>4) ? 4 : balance_iteration;
    // store the main proxy
    mainProxy = thisProxy;

    // print info
    CkPrintf("Running on %d processors with %d chares per pe\n", 
	     CkNumPes(), num_chare_blocks*4);

    // Create new array of worker chares. The element constructors will
    // contact this object to start the computation.
    arrayProxy = CProxy_LB_Test::ckNew(num_chare_blocks*4*CkNumPes());

  }

  void report_in(CkReductionMsg *msg) {
    delete msg;
    workStartTimestamp = CkWallTimer();
    CkPrintf("All array elements ready at %f seconds. Computation Begins\n",
	     workStartTimestamp - timestamp);
    arrayProxy.next_iter();
  }

  // Reduction callback client
  void iterBarrier(CkReductionMsg *msg) {
    totalChareWorkTime += ((double *) msg->getData())[0];
    delete msg;
    arrayProxy.next_iter();
  }

  // Each worker reports back to here when it completes all work
  void report_done(CkReductionMsg *msg) {
    delete msg;
    CkPrintf("Total work performed = %f seconds\n", totalChareWorkTime);

    CkPrintf("Average total chare work per iteration = %f seconds\n",
	     totalChareWorkTime/total_iterations);
    CkPrintf("Average iteration time = %f seconds\n",
	     (CkWallTimer() - workStartTimestamp)/total_iterations);
    CkPrintf("Done after %f seconds\n", CkWallTimer() - timestamp);
    CkExit();
  }
};

class LB_Test: public CBase_LB_Test {
public:
  int iteration;
  int work_factor;
  double timeTaken;
  double *a, *b, *c;
  // Initialization and start of iterations. 
  // Chares starting on even-numbered processors get a single work
  // unit. 
  // Chares starting on odd-numbered processors get two work units.
  // NOTE: These work factors do not change after migration! That's
  //       the point of this example!
  LB_Test() {
    
    if (CkMyPe() % 2 == 0) {
      work_factor = 1;
    } else {
      work_factor = 2;
    }
    work_factor*=workWeight;
    iteration = 0;
    timeTaken = 0.0;
    usesAtSync = CmiTrue;


    a = new double[arraySize];
    b = new double[arraySize];
    c = new double[arraySize];
    for(int i=0;i<arraySize;i++){
      a[i] = (double) thisIndex;
      b[i] = (double) thisIndex;
      c[i] = (double) thisIndex;
    }
    CkCallback cb(CkIndex_Main::report_in(NULL),mainProxy);
    int one=1;
    contribute(sizeof(int),&one,CkReduction::sum_int,cb);
  }

  // For migration
  LB_Test(CkMigrateMessage* m) {
  }

  // Destructor
  ~LB_Test() {
    delete [] a;
    delete [] b;
    delete [] c;
  }

  // Load Balancing happens at lower of 4 iterations or one quarter mark
  void next_iter() {
    if (iteration < total_iterations) {
      if ((iteration == balance_iteration) && usesAtSync) {
	AtSync();
      } else {
	compute();
      }
    } else {
      CkCallback cb(CkIndex_Main::report_done(NULL),mainProxy);
      contribute(sizeof(int),&iteration,CkReduction::sum_int,cb);
    }
  }

  void compute() {

    double timeStamp = CkWallTimer();
    double checksumC=0.0;
    for(int j=0;j<500*work_factor;j++){
      for(int i=0;i<arraySize;i++){
	a[i] = (double) thisIndex+j;
	b[i] = (double) (thisIndex+j)*3.0;
      }
      for(int i=0;i<arraySize/2;i++){
	c[i] = a[2*i]*b[2*i]*a[i];
	c[2*i] = a[2*i];
      }
      for(int i=0;i<arraySize;i++)
	checksumC+=c[i];
    }


    timeTaken = CkWallTimer() - timeStamp;
    // Sanity output
    if (((iteration == 0) || (iteration == total_iterations-1)) &&
	((thisIndex == 0) || (thisIndex == num_chare_blocks*4))) {
      CkPrintf("[%d] Element %d took %f seconds for work %d at iteration %d sumC %0.4g\n", 
	       CkMyPe(), thisIndex, timeTaken,work_factor,iteration,checksumC);
    }
    iteration++;
    CkCallback cb(CkIndex_Main::iterBarrier(NULL),mainProxy);
    contribute(sizeof(double),&timeTaken,CkReduction::sum_double,cb);

  }
  
  void ResumeFromSync(void) { // Called by Load-balancing framework
    compute();
  }
  
  void pup(PUP::er &p)
  {
    CBase_LB_Test::pup(p);
    p | iteration;
    p | work_factor;
    p | arraySize;
    p | timeTaken;
    if(p.isUnpacking())
      {
	a=new double[arraySize];
	b=new double[arraySize];
	c=new double[arraySize];
      }
    PUParray(p,a,arraySize); 
    PUParray(p,b,arraySize); 
    PUParray(p,c,arraySize); 
  }
};

#include "LB_Test.def.h"

